import java.awt.*;
import java.awt.event.*;
import java.awt.image.ImageObserver;
import java.awt.image.ImageConsumer;
import java.awt.image.ColorModel;
import java.awt.image.BufferedImage;
import java.awt.image.RenderedImage;
import javax.swing.*;
import javax.swing.filechooser.FileFilter;
import javax.imageio.ImageIO;
import javax.imageio.ImageWriter;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.util.List;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Hashtable;
import java.util.Arrays;
import java.util.Collections;
import java.util.Iterator;
import java.util.Comparator;

/**
 * @author David Foster <davidfstr@gmail.com>
 */
public class ImageFilterDemo {
	// ### FILTER CONFIGURATION ###
	
	/**
	 * contains the filters that are displayed in the UI.
	 * 
	 * Supported filter types include:
	 * <ul>
	 * <li>ImageFilter</li>
	 * <li>BufferedImageFilter</li>
	 * <li>BufferedImageFilter2</li> (since v1.2.2)
	 * <li>PixelBufferToBufferedImageFilter</li> (since v1.2.2)
	 * <li>PixelBufferFilter</li>
	 * <li>PixelBufferFilter2</li> (since v1.0.1)
	 * <li>PixelFilter</li>
	 * </ul>
	 */
	private static List filterList;		// List<SUPPORTED_FILTER_TYPE>
	
	static {
		/* Initialize filter list */
		{
			filterList = new ArrayList();
			
			List filterPluginClasses = ImageFilterDemoPlugins.filterPluginClasses;
			for (int i=0, n=filterPluginClasses.size(); i<n; i++) {
				Class curFilterPluginClass = (Class) filterPluginClasses.get(i);
				
				Object curFilterPluginInstance;
				try {
					curFilterPluginInstance = curFilterPluginClass.newInstance();
				} catch (IllegalAccessException e) {
					System.err.println(
						"Unable to instantiate filter \""+curFilterPluginClass.getName()+"\": "+
						"Either the class or its nullary constructor is not accessible");
					continue;
				} catch (InstantiationException e) {
					System.err.println(
						"Unable to instantiate filter \""+curFilterPluginClass.getName()+"\": "+
						"Either the class has no nullary constructor, is an abstract class, or is an interface");
					continue;
				} catch (ExceptionInInitializerError e) {
					System.err.println(
						"Unable to instantiate filter \""+curFilterPluginClass.getName()+"\": "+
						"Exception was thrown by nullary constructor:");
					e.printStackTrace(System.err);
					continue;
				} catch (SecurityException e) {
					System.err.println(
						"Unable to instantiate filter \""+curFilterPluginClass.getName()+"\": "+
						"Insufficient permissions:");
					e.printStackTrace(System.err);
					continue;
				} catch (Throwable e) {
					System.err.println(
						"Unable to instantiate filter \""+curFilterPluginClass.getName()+"\": "+
						"Unknown error:");
					e.printStackTrace(System.err);
					continue;
				}
				
				filterList.add(curFilterPluginInstance);
			}
			
			// Arrange the filter list so as to appear in alphabetical order in the GUI
			Collections.sort(filterList, new Comparator() {
				public int compare(Object o1, Object o2) {
					String guiName1 = o1.toString();
					String guiName2 = o2.toString();
					return String.CASE_INSENSITIVE_ORDER.compare(guiName1, guiName2);
				}
			});
		}
	}
	
	// ### MAIN ###
	
	private static JFrame mainFrame;
	
	private static JLabel fileLineLabel;
	private static JFileWell inputFileWell;
	private static JButton chooseFileButton;
	
	private static JLabel filterLineLabel;
	private static JComboBox filterComboBox;
	private static JButton optionsButton;
	
	private static ComponentWell inputImageView;
	private static ComponentWell outputImageView;
	
	private static JButton playPauseButton;
	private static JButton saveOutputImageButton;
	
	private static Image inputImage;
	private static Image outputImage;
	
	private static boolean playingImageSequence;
	private static LinkedList playbackFileList = new LinkedList();	// List<File>
	private static boolean playbackFileListCached = false;
	
	public static void main(String[] args) {
		inputFileWell = new JFileWell();
		inputFileWell.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				// Update UI to reflect the new input file
				inputFileChanged();
			}
		});
		
		chooseFileButton = new JButton("Choose...");
		chooseFileButton.setFont(JFileWell.FONT);
		chooseFileButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				JFileChooser fileChooser = new JFileChooser();
				
				/* Load/select the last input file that was chosen */
				File lastSelectedInputFile = ImageFilterDemoPrefs.getLastSelectedInputFile();
				tryToSelectFile(fileChooser, lastSelectedInputFile);
				
				int chooseResult = fileChooser.showOpenDialog(mainFrame);
				if (chooseResult == JFileChooser.APPROVE_OPTION) {
					File selectedFile = fileChooser.getSelectedFile();
					
					/* Save the identity of the file that was selected */
					ImageFilterDemoPrefs.setLastSelectedInputFile(selectedFile);
					
					// Designate the selected file to be the input file
					setInputFile(selectedFile);
				}
			}
		});
		
		FilterComboBoxModel filterComboBoxModel = new FilterComboBoxModel();
		filterComboBox = new JComboBox(filterComboBoxModel);
		filterComboBox.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				// Update Output pane
				filterChanged();
			}
		});
		
		optionsButton = new JButton("Options");
		optionsButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				if (!optionsDrawer.isVisible()) {
					showOptionsDrawer();
				} else {
					hideOptionsDrawer();
				}
			}
		});
		
		playPauseButton = new JButton();
		playPauseButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				playPauseButtonWasClicked();
			}
		});
		
		saveOutputImageButton = new JButton("Save...");
		saveOutputImageButton.setEnabled(false);
		saveOutputImageButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				saveOutputButtonWasClicked();
			}
		});
		
		inputImageView = new ComponentWell(null);
		outputImageView = new ComponentWell(null);
		
		initMainFrame();
		initOptionsDrawer();		// requires main frame to be initialized
		
		/* Generate events to update UI to reflect initial state */
		updateInputPane();
		updateOutputPane();
		setPlaybackEnabled(false);
		setPlaying(false);
		
		/* Show main window */
		mainFrame.pack();
		mainFrame.show();
	}
	
	private static void initMainFrame() {
		Container fileLine = new Container();
		fileLine.setLayout(new ListLayout(ListLayout.X_AXIS, ListLayout.CENTER));
		fileLine.add(makeBold(fileLineLabel = new JLabel("File: ")));
		fileLine.add(inputFileWell, Boolean.TRUE);
		fileLine.add(chooseFileButton);
		
		JPanel inputPane = new JPanel();
		inputPane.setBorder(BorderFactory.createTitledBorder("Input"));
		inputPane.setLayout(new ListLayout(ListLayout.Y_AXIS, ListLayout.FILL));
		inputPane.add(fileLine);
		inputPane.add(inputImageView, Boolean.TRUE);
		
		Container filterLine = new Container();
		filterLine.setLayout(new ListLayout(ListLayout.X_AXIS, ListLayout.CENTER));
		filterLine.add(makeBold(filterLineLabel = new JLabel("Filter: ")));
		filterLine.add(filterComboBox, Boolean.TRUE);
		filterLine.add(optionsButton);
		
		JPanel outputPane = new JPanel();
		outputPane.setBorder(BorderFactory.createTitledBorder("Output"));
		outputPane.setLayout(new ListLayout(ListLayout.Y_AXIS, ListLayout.FILL));
		outputPane.add(filterLine);
		outputPane.add(outputImageView, Boolean.TRUE);
		
		Container mainPane = new Container();
		mainPane.setLayout(new GridLayout(1, 2));
		mainPane.add(inputPane);
		mainPane.add(outputPane);
		
		Container statusBar = new Container();
		statusBar.setLayout(new ListLayout(ListLayout.X_AXIS, ListLayout.CENTER));
		statusBar.add(playPauseButton);
		statusBar.add(Box.createHorizontalGlue(), Boolean.TRUE);
		statusBar.add(saveOutputImageButton);
		
		Container content = new Container();
		content.setLayout(new ListLayout(ListLayout.Y_AXIS, ListLayout.FILL));
		content.add(mainPane, Boolean.TRUE);
		content.add(statusBar);
		
		mainFrame = new JFrame();
		mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		mainFrame.setContentPane(content);
	}
	
	private static JLabel makeBold(JLabel label) {
		label.setFont(label.getFont().deriveFont(Font.BOLD));
		return label;
	}
	
	// ### IMAGE SEQUENCE PLAYBACK ###
	
	// NOTE: must only be called from the event dispatching thread
	private static void playPauseButtonWasClicked() {
		if (!isPlaying()) {
			playButtonWasClicked();
		} else {
			pauseButtonWasClicked();
		}
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void playButtonWasClicked() {
		/* 
		 * Make sure a filter has been selected;
		 * if this is not the case, tell the user
		 */
		if (getSelectedFilter() == null) {
			JOptionPane.showMessageDialog(
				mainFrame,
				"No filter has been selected.",
				"Animation Error",
				JOptionPane.ERROR_MESSAGE,
				null);
			return;
		}
		
		/* 
		 * Load playback-file-list with every file in the same
		 * directory as the current input file whose name comes
		 * alphabetically after the current input file's name.
		 */
		if (!playbackFileListCached) {
			File inputFile = getInputFile();
			File parentDir = inputFile.getParentFile();
			if (parentDir != null) {
				// Get the list of names of the files in the same
				// directory as the input file, in alphabetical order
				String[] filenames = parentDir.list();
				Arrays.sort(filenames, String.CASE_INSENSITIVE_ORDER);
				
				// Find the input file's name in the list of filenames
				// in the same directory
				try {
					String canonicalInputFileName = inputFile.getCanonicalFile().getName();
					int off = Arrays.binarySearch(
						filenames,
						canonicalInputFileName,
						String.CASE_INSENSITIVE_ORDER);
					if (off >= 0) {
						// Populate the playback-file-list
						for (int i=off+1, n=filenames.length; i<n; i++) {
							playbackFileList.add(new File(parentDir, filenames[i]));
						}
					}
				} catch (IOException e) {
					// ignore I/O errors
				}
			}
			
			playbackFileListCached = true;
		}
		
		/* Initiate playback if there are any more images to play in the sequence */
		if (!playbackFileList.isEmpty()) {
			setPlaying(true);
			
			// Trigger retrieval of second frame of animation
			outputImageChanged();
		}
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void pauseButtonWasClicked() {
		// FIXME: implement
		setPlaying(false);
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void setPlaybackEnabled(boolean enabled) {
		playPauseButton.setEnabled(enabled);
		
		if (!enabled) {
			setPlaying(false);
		}
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void setPlaying(boolean enabled) {
		playingImageSequence = enabled;
		
		/* Update UI: Play/Pause Button */
		if (!playingImageSequence) {
			playPauseButton.setText("Play");
		} else {
			playPauseButton.setText("Pause");
		}
		
		/* Update UI: Input/Filter Settings Enabled */
		fileLineLabel.setEnabled(!playingImageSequence);
		inputFileWell.setEnabled(!playingImageSequence);
		chooseFileButton.setEnabled(!playingImageSequence);
		
		filterLineLabel.setEnabled(!playingImageSequence);
		filterComboBox.setEnabled(!playingImageSequence);
		optionsButton.setEnabled(!playingImageSequence);
		
		saveOutputImageButton.setEnabled(!playingImageSequence);
	}
	
	private static boolean isPlaying() {
		return playingImageSequence;
	}
	
	private static void dumpPlaybackFileList() {
		playbackFileList.clear();
		playbackFileListCached = false;
	}
	
	// ### OPTIONS DRAWER ###
	
	private static JWindow optionsDrawer;
	private static ComponentWell optionDrawerView;
	
	private final static Component OPTIONS_ERROR_FILTER_NOT_SELECTED =
		new JDesktopMessagePane("   No filter selected   ");
	private final static Component OPTIONS_STATUS_NO_PARAMS =
		new JDesktopMessagePane("   No parameters   ");
	
	private static void initOptionsDrawer() {
		optionDrawerView = new ComponentWell(null);
		
		// Ensure that the position/size of the Options drawer 
		// is updated as the main frame is moved/resized
		// Mac OS X: doesn't work very well *while* the main frame is being dragged
		mainFrame.addComponentListener(new ComponentAdapter() {
			public void componentMoved(ComponentEvent evt) {
				if (optionsDrawer.isVisible()) {
					reshapeOptionsDrawer();
				}
			}
			
			public void componentResized(ComponentEvent evt) {
				if (optionsDrawer.isVisible()) {
					reshapeOptionsDrawer();
				}
			}
		});
		
		if (mainFrame == null)
			throw new IllegalStateException("mainFrame is not yet initialized");
		optionsDrawer = new JWindow(mainFrame);
		optionsDrawer.setContentPane(optionDrawerView);
		
		updateOptionsDrawer();
	}
	
	private static void showOptionsDrawer() {
		reshapeOptionsDrawer();
		optionsDrawer.show();
	}
	
	private static void hideOptionsDrawer() {
		// Hide the options drawer
		optionsDrawer.hide();
	}
	
	private static void reshapeOptionsDrawer() {
		/* Calculate the appropriate size and position of the options drawer */
		optionsDrawer.pack();
		int width = optionsDrawer.getWidth();
		int height = Math.min(optionsDrawer.getHeight(), mainFrame.getHeight());
		int x = mainFrame.getX() + mainFrame.getWidth();
		int y = mainFrame.getY() + (mainFrame.getHeight() - height)/2;
		
		/* Reshape the options drawer appropriately */
		optionsDrawer.setLocation(x, y);
		optionsDrawer.setSize(width, height);
	}
	
	private final static FilterParam[] NO_FILTER_PARAMS = new FilterParam[0];
	private final static int OPTION_DRAWER_PADDING = 5;
	
	private static void updateOptionsDrawer() {
		Object selectedFilter = getSelectedFilter();
		
		if (selectedFilter == null) {
			optionDrawerView.setComponent(OPTIONS_ERROR_FILTER_NOT_SELECTED);
		} else {
			FilterParam[] params;
			if (selectedFilter instanceof ParameterizedFilter) {
				params = ((ParameterizedFilter) selectedFilter).getParams();
				if (params == null) {
					params = NO_FILTER_PARAMS;
				}
			} else {
				params = NO_FILTER_PARAMS;
			}
			
			if (params.length != 0) {
				JPanel content = new JPanel();
				content.setBorder(BorderFactory.createEmptyBorder(
					OPTION_DRAWER_PADDING, OPTION_DRAWER_PADDING,
					OPTION_DRAWER_PADDING, OPTION_DRAWER_PADDING));
				GridBagLayout g = new GridBagLayout();
				content.setLayout(g);
				
				GridBagConstraints c = new GridBagConstraints();
				c.fill = GridBagConstraints.BOTH;
				
				for (int i=0, n=params.length; i<n; i++) {
					FilterParam curParam = params[i];
					
					JLabel paramNameLabel = new JLabel(curParam.getName()+": ");
					c.gridx = 0;
					c.gridy = i;
					g.setConstraints(paramNameLabel, c);
					content.add(paramNameLabel);
					
					Component paramValueEditorComponent = createEditorComponentForValueOfFilterParam(curParam);
					c.gridx = 1;
					c.gridy = i;
					g.setConstraints(paramValueEditorComponent, c);
					content.add(paramValueEditorComponent);
				}
				
				JScrollPane scrollingContent = new JScrollPane(content);
				
				optionDrawerView.setComponent(scrollingContent);
				reshapeOptionsDrawer();
			} else {
				optionDrawerView.setComponent(OPTIONS_STATUS_NO_PARAMS);
				reshapeOptionsDrawer();
			}
		}
	}
	
	private static Component createEditorComponentForValueOfFilterParam(FilterParam param) {
		if (param instanceof IntFilterParam) {
			final IntFilterParam intParam = (IntFilterParam) param;
			final JTextField valueView = new JTextField(String.valueOf(intParam.getValue()), 4);
			
			final Runnable valueUpdater = new Runnable() {
				String lastTextOfValueView = valueView.getText();
				
				// NOTE: must only be called on the event dispatching thread
				public void run() {
					try {
						int proposedNewValue = Integer.parseInt(valueView.getText());
						
						int actualNewValue = proposedNewValue;
						if (actualNewValue < intParam.getMin())
							actualNewValue = intParam.getMin();
						if (actualNewValue > intParam.getMax())
							actualNewValue = intParam.getMax();
						
						// Update UI
						if (proposedNewValue != actualNewValue) {
							// Typed value is out of range; make it in range
							valueView.setText(String.valueOf(actualNewValue));
							valueView.selectAll();
						}
						lastTextOfValueView = valueView.getText();
						
						// Update param
						intParam.setValue(actualNewValue);
						
						// Update output image
						filterParamsChanged();
					} catch (NumberFormatException e) {
						// Unable to parse the typed value - use the old value
						valueView.setText(lastTextOfValueView);
						valueView.selectAll();
					}
					
				}
			};
			
			// Update value when user presses enter while editing
			valueView.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent evt) {
					valueUpdater.run();
				}
			});
			
			// Update value when the value-editor loses focus
			valueView.addFocusListener(new FocusListener() {
				public void focusGained(FocusEvent evt) {
					valueView.selectAll();
				}
				
				public void focusLost(FocusEvent evt) {
					valueUpdater.run();
				}
			});
			
			return valueView;
		}
		
		// XXX: code very similar to that which generates the editor component for IntFilterParams
		if (param instanceof FloatFilterParam) {
			final FloatFilterParam floatParam = (FloatFilterParam) param;
			final JTextField valueView = new JTextField(String.valueOf(floatParam.getValue()), 4);
			
			final Runnable valueUpdater = new Runnable() {
				String lastTextOfValueView = valueView.getText();
				
				// NOTE: must only be called on the event dispatching thread
				public void run() {
					try {
						float proposedNewValue = Float.parseFloat(valueView.getText());
						
						float actualNewValue = proposedNewValue;
						if (actualNewValue < floatParam.getMin())
							actualNewValue = floatParam.getMin();
						if (actualNewValue > floatParam.getMax())
							actualNewValue = floatParam.getMax();
						
						// Update UI
						if (proposedNewValue != actualNewValue) {
							// Typed value is out of range; make it in range
							valueView.setText(String.valueOf(actualNewValue));
							valueView.selectAll();
						}
						lastTextOfValueView = valueView.getText();
						
						// Update param
						floatParam.setValue(actualNewValue);
						
						// Update output image
						filterParamsChanged();
					} catch (NumberFormatException e) {
						// Unable to parse the typed value - use the old value
						valueView.setText(lastTextOfValueView);
						valueView.selectAll();
					}
				}
			};
			
			// Update value when user presses enter while editing
			valueView.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent evt) {
					valueUpdater.run();
				}
			});
			
			// Update value when the value-editor loses focus
			valueView.addFocusListener(new FocusListener() {
				public void focusGained(FocusEvent evt) {
					valueView.selectAll();
				}
				
				public void focusLost(FocusEvent evt) {
					valueUpdater.run();
				}
			});
			
			return valueView;
		}
		
		// XXX: code very similar to that which generates the editor component for IntFilterParams
		if (param instanceof DoubleFilterParam) {
			final DoubleFilterParam doubleParam = (DoubleFilterParam) param;
			final JTextField valueView = new JTextField(String.valueOf(doubleParam.getValue()), 4);
			
			final Runnable valueUpdater = new Runnable() {
				String lastTextOfValueView = valueView.getText();
				
				// NOTE: must only be called on the event dispatching thread
				public void run() {
					try {
						double proposedNewValue = Double.parseDouble(valueView.getText());
						
						double actualNewValue = proposedNewValue;
						if (actualNewValue < doubleParam.getMin())
							actualNewValue = doubleParam.getMin();
						if (actualNewValue > doubleParam.getMax())
							actualNewValue = doubleParam.getMax();
						
						// Update UI
						if (proposedNewValue != actualNewValue) {
							// Typed value is out of range; make it in range
							valueView.setText(String.valueOf(actualNewValue));
							valueView.selectAll();
						}
						lastTextOfValueView = valueView.getText();
						
						// Update param
						doubleParam.setValue(actualNewValue);
						
						// Update output image
						filterParamsChanged();
					} catch (NumberFormatException e) {
						// Unable to parse the typed value - use the old value
						valueView.setText(lastTextOfValueView);
						valueView.selectAll();
					}
				}
			};
			
			// Update value when user presses enter while editing
			valueView.addActionListener(new ActionListener() {
				public void actionPerformed(ActionEvent evt) {
					valueUpdater.run();
				}
			});
			
			// Update value when the value-editor loses focus
			valueView.addFocusListener(new FocusListener() {
				public void focusGained(FocusEvent evt) {
					valueView.selectAll();
				}
				
				public void focusLost(FocusEvent evt) {
					valueUpdater.run();
				}
			});
			
			return valueView;
		}
		
		if (param instanceof BooleanFilterParam) {
			final BooleanFilterParam booleanParam = (BooleanFilterParam) param;
			final JCheckBox valueView = new JCheckBox("", booleanParam.getValue());
			valueView.addItemListener(new ItemListener() {
				public void itemStateChanged(ItemEvent evt) {
					booleanParam.setValue(valueView.isSelected());
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							filterParamsChanged();
						}
					});
				}
			});
			return valueView;
			
			//return new JLabel("not yet implemented");
		}
		
		throw new IllegalArgumentException("unknown type of FilterParam: "+param.getClass().getName());
	}
	
	// ### EXPORTING THE OUTPUT IMAGE ###
	
	private static void saveOutputButtonWasClicked() {
		/* 
		 * Get a list of format names that we support
		 * for exporting, in alphabetical order
		 */
		List writerFormatNames = new ArrayList(Arrays.asList(		// List<String>
			ImageIO.getWriterFormatNames()));
		Collections.sort(writerFormatNames, String.CASE_INSENSITIVE_ORDER);
		
		/* Create FileFilters for each supported export format */
		ImageWriterFormatNameFileFilter[] fileFilters = 
			new ImageWriterFormatNameFileFilter[writerFormatNames.size()];
		for (int i=0, n=writerFormatNames.size(); i<n; i++) {
			String curFormatName = (String) writerFormatNames.get(i);
			fileFilters[i] = new ImageWriterFormatNameFileFilter(curFormatName);
		}
		
		JFileChooser fileChooser = new JFileChooser();
		
		// Remove preexisting file filters from file chooser
		{
			FileFilter[] origFileFilters = fileChooser.getChoosableFileFilters();
			for (int i=0, n=origFileFilters.length; i<n; i++) {
				fileChooser.removeChoosableFileFilter(origFileFilters[i]);
			}
		}
		
		// Add custom file filters to file chooser
		for (int i=0, n=fileFilters.length; i<n; i++) {
			fileChooser.addChoosableFileFilter(fileFilters[i]);
		}
		
		/* Load the parent directory containing of the last output file that was selected */
		{
			File lastSelectedOutputDir = ImageFilterDemoPrefs.getLastSelectedOutputDirectory();
			fileChooser.setCurrentDirectory(lastSelectedOutputDir);
		}
		
		/* Load the last output format that was chosen */
		{
			String lastChosenOutputFormatName = ImageFilterDemoPrefs.getLastChosenOutputFormatName();
			if (lastChosenOutputFormatName != null) {
				ImageWriterFormatNameFileFilter lastChosenOutputFilter = null;
				for (int i=0, n=fileFilters.length; i<n; i++) {
					ImageWriterFormatNameFileFilter curFileFilter = fileFilters[i];
					if (curFileFilter.getFormatName().equals(lastChosenOutputFormatName)) {
						lastChosenOutputFilter = curFileFilter;
						break;
					}
				}
				
				if (lastChosenOutputFilter != null) {
					fileChooser.setFileFilter(lastChosenOutputFilter);
				} else {
					// Unable to find the last chosen output filter; use the default
				}
			} else {
				// Last chosen output filter is undefined; use the default
			}
		}
		
		int chooseResult = fileChooser.showSaveDialog(mainFrame);
		if (chooseResult == JFileChooser.APPROVE_OPTION) {
			File selectedFile = fileChooser.getSelectedFile();
			String chosenFormatName = 
				((ImageWriterFormatNameFileFilter) fileChooser.getFileFilter()).getFormatName();
			
			/* Save the identity of the file that was selected */
			ImageFilterDemoPrefs.setLastSelectedOutputDirectory(selectedFile.getParentFile());
			
			/* Save the output format that was chosen */
			ImageFilterDemoPrefs.setLastChosenOutputFormatName(chosenFormatName);
			
			/* Save the output image to the file */
			RenderedImage imageToSave;
			if (outputImage instanceof RenderedImage) {
				imageToSave = (RenderedImage) outputImage;
			} else {
				BufferedImage outputImageCopy;
				{
					outputImageCopy = new BufferedImage(
						outputImage.getWidth((ImageObserver) null),
						outputImage.getHeight((ImageObserver) null),
						BufferedImage.TYPE_INT_ARGB);
					outputImageCopy.getGraphics().drawImage(
						outputImage,
						0, 0,
						(ImageObserver) null);
				}
				imageToSave = outputImageCopy;
			}
			
			try {
				ImageIO.write(
					(RenderedImage) imageToSave,
					chosenFormatName,
					selectedFile);
			} catch (IOException e) {
				System.err.println("Error while saving output image:");
				e.printStackTrace(System.err);
				
				JOptionPane.showMessageDialog(
					mainFrame,
					"An I/O error occurred while saving the output image: "+e,
					"I/O Error",
					JOptionPane.ERROR_MESSAGE,
					null);
				return;
			}	
		}
	}
	
	private static class ImageWriterFormatNameFileFilter extends FileFilter {
		private String formatName;
		
		// cached, lowercase, with initial period
		private static List supportedFileSuffixes = null;
		
		// ### INIT ###
		
		public ImageWriterFormatNameFileFilter(String formatName) {
			this.formatName = formatName;
		}
		
		// ### FILE FILTER ###
		
		public String getDescription() {
			// TODO: provide better descriptions than just the short format name, if possible
			return this.formatName;
		}
		
		public boolean accept(File file) {
			if (supportedFileSuffixes == null) {
				supportedFileSuffixes = new ArrayList();
				{
					Iterator imageWriterIter = ImageIO.getImageWritersByFormatName(this.formatName);
					while (imageWriterIter.hasNext()) {
						ImageWriter curImageWriter = (ImageWriter) imageWriterIter.next();
						supportedFileSuffixes.addAll(
							Arrays.asList(
								curImageWriter.getOriginatingProvider().getFileSuffixes()));
					}
					
					// Make all suffixes lowercase and prepend a period to each suffix
					for (int i=0, n=supportedFileSuffixes.size(); i<n; i++) {
						String curSuffix = (String) supportedFileSuffixes.get(i);
						supportedFileSuffixes.set(
							i,
							"."+curSuffix.toLowerCase());
					}
				}
			}
			
			/* Accept/reject the file based on whether its name ends with a supported suffix */
			String lowercaseFileName = file.getName().toLowerCase();
			for (int i=0, n=supportedFileSuffixes.size(); i<n; i++) {
				String curSuffix = (String) supportedFileSuffixes.get(i);
				if (lowercaseFileName.endsWith(curSuffix)) {
					return true;
				}
			}
			return false;
		}
		
		// ### ACCESSORS ###
		
		public String getFormatName() {
			return this.formatName;
		}
	}
	
	// ### INTERNAL EVENTS ###
	
	// NOTE: must only be called from the event dispatching thread
	private static void inputFileChanged() {
		updateInputPane();
		
		/* Update playback controls */
		setPlaybackEnabled(getInputFile() != null);
		if (!isPlaying()) {
			// Uncache the playback-file-list
			dumpPlaybackFileList();
		}
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void inputImageChanged() {
		updateOutputPane();
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void filterChanged() {
		updateOutputPane();
		updateOptionsDrawer();
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void filterParamsChanged() {
		updateOutputPane();
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void outputImageChanged() {
		if (!isPlaying()) {
			saveOutputImageButton.setEnabled(outputImage != null);
		} else {
			if (outputImage != null) {	// ensure that we really have a new output image
				/* Trigger retrieval of the next frame in the animation (if there is one) */
				if (playbackFileList.isEmpty()) {
					/* No more frames left */
					setPlaying(false);
				} else {
					/* More frames remain - load up the next frame */
					File nextInputFile = (File) playbackFileList.removeFirst();
					setInputFile(nextInputFile);
				}
			}
		}
	}
	
	// ### GUI UPDATES ###
	
	private static Component INPUT_ERROR_FILE_NOT_SELECTED =
		new JDesktopMessagePane("   No image file selected   ");
	
	private static Component INPUT_ERROR_BAD_IMAGE =
		new JDesktopMessagePane("   Unable to load image!   ");
	
	private static Component INPUT_STATUS_LOADING_IMAGE =
		new JDesktopMessagePane("   Loading image...   ");
	
	private static Component OUTPUT_STATUS_WAITING_FOR_INPUT =
		new JDesktopMessagePane("   Waiting for input image   ");
	
	private static Component OUTPUT_STATUS_APPLYING_FILTER =
		new JDesktopMessagePane("   Applying filter...   ");
	
	private static Component OUTPUT_ERROR_FILTER_NOT_SELECTED =
		new JDesktopMessagePane("   No filter selected   ");
	
	// NOTE: must only be called from the event dispatching thread
	private static void updateInputPane() {
		// Unload image from old input file (and update Output pane)
		inputImage = null;
		inputImageChanged();
		
		final File newInputFile = getInputFile();
		if (newInputFile == null) {
			inputImageView.setComponent(INPUT_ERROR_FILE_NOT_SELECTED);
		} else {
			if (!isPlaying()) {		// suppress loading screens when playing animation
				inputImageView.setComponent(INPUT_STATUS_LOADING_IMAGE);
			}
			
			/* Load the image asynchronously in a new thread */
			chooseFileButton.setEnabled(false);
			new Thread("Image Loader") {
				public void run() {
					/* Load the image */
					BufferedImage newInputImage;
					boolean inputImageLoadedSuccessfully;
					try {
						newInputImage = ImageIO.read(newInputFile);
						if (newInputImage == null) {
							System.err.println("Format of input image is not supported");
							inputImageLoadedSuccessfully = false;
						} else {
							inputImageLoadedSuccessfully = true;
						}
					} catch (IOException e) {
						System.err.println("I/O error while loading input image:");
						e.printStackTrace(System.err);
						
						newInputImage = null;
						inputImageLoadedSuccessfully = false;
					}
					
					/* Update the UI */
					final BufferedImage newInputImage_final = newInputImage;
					final boolean inputImageLoadedSuccessfully_final = inputImageLoadedSuccessfully;
					SwingUtilities.invokeLater(new Runnable() {
						public void run() {
							if (inputImageLoadedSuccessfully_final) {
								// Update UI to display image (as it is loaded)
								inputImageView.setComponent(new JLabel(new ImageIcon(newInputImage_final)));
								mainFrame.pack();		// update window size
								
								// Remember the loaded image
								inputImage = newInputImage_final;
								inputImageChanged();
							} else {
								// An error occurred while loading the image
								inputImageView.setComponent(INPUT_ERROR_BAD_IMAGE);
							}
							chooseFileButton.setEnabled(true && (!isPlaying()));
						}
					});
				}
			}.start();
		}
	}
	
	// NOTE: must only be called from the event dispatching thread
	private static void updateOutputPane() {
		// Unload old output image
		outputImage = null;
		outputImageChanged();
		
		final ImageFilter selectedImageFilter = getSelectedFilterAsImageFilter();
		if (selectedImageFilter == null) {
			outputImageView.setComponent(OUTPUT_ERROR_FILTER_NOT_SELECTED);
		} else {
			if (inputImage == null) {
				if (!isPlaying()) {		// suppress loading screens when playing animation
					outputImageView.setComponent(OUTPUT_STATUS_WAITING_FOR_INPUT);
				}
			} else {
				if (!isPlaying()) {		// suppress loading screens when playing animation
					outputImageView.setComponent(OUTPUT_STATUS_APPLYING_FILTER);
				}
				
				new Thread("Image Filter") {
					public void run() {
						final Image filteredImage = selectedImageFilter.createFilteredImage(inputImage);
						SwingUtilities.invokeLater(new Runnable() {
							public void run() {
								// Display the filtered image
								outputImageView.setComponent(new JLabel(new ImageIcon(filteredImage)));
								mainFrame.pack();		// update window size
								
								// Remember the filtered image
								outputImage = filteredImage;
								outputImageChanged();
							}
						});
					}
				}.start();
			}
		}
	}
	
	// ### INTERNAL ACCESSORS ###
	
	// NOTE: must only be called from the event dispatching thread
	public static void setInputFile(File inputFile) {
		inputFileWell.setFile(inputFile);
	}
	
	public static File getInputFile() {
		return inputFileWell.getFile();
	}
	
	public static Object getSelectedFilter() {
		return filterComboBox.getSelectedItem();
	}
	
	public static ImageFilter getSelectedFilterAsImageFilter() {
		return filterAsImageFilter(getSelectedFilter());
	}
	
	// ### UTILITY ###
	
	/**
	 * Tries to select the specified <tt>file</tt> in the <tt>fileChooser</tt>.
	 * <p>
	 * It <tt>file</tt> is <tt>null</tt>, the file chooser's selected file and
	 * current directory will not be changed.
	 */
	private static void tryToSelectFile(JFileChooser fileChooser, File file) {
		if (file != null) {
			/*
			 * Try to select the file that was specified
			 */
			if (file.exists()) {
				// Select the file that was specified
				fileChooser.setSelectedFile(file);
			} else {
				// If the specified file could not be found,
				// try to set the current directory to the directory
				// that contained (or that is "closest" to) the file
				File dirContainingLastSelectedInputFile = file.getParentFile();
				fileChooser.setCurrentDirectory(dirContainingLastSelectedInputFile);
			}
		} else {
			/* 
			 * Let the file chooser use the default
			 * current directory and selected file
			 */
		}
	}
	
	// ### IMAGE FILTER CONVERSION ###
	
	/**
	 * Converts a <tt>filter</tt> of one of the supported filter types
	 * to an <tt>ImageFilter</tt>.
	 */
	private static ImageFilter filterAsImageFilter(Object filter) {
		if (filter == null)
			return null;
		
		if (filter instanceof PixelFilter)
			filter = new PixelFilterAsPixelBufferFilter((PixelFilter) filter);
		if (filter instanceof PixelBufferFilter)
			filter = new PixelBufferFilterAsBufferedImageFilter((PixelBufferFilter) filter);
		if (filter instanceof PixelBufferFilter2)
			filter = new PixelBufferFilter2AsBufferedImageFilter((PixelBufferFilter2) filter);
		if (filter instanceof PixelBufferToBufferedImageFilter)
			filter = new PixelBufferToBufferedImageFilterAsBufferedImageFilter2((PixelBufferToBufferedImageFilter) filter);
		if (filter instanceof BufferedImageFilter)
			filter = new BufferedImageFilterAsImageFilter((BufferedImageFilter) filter);
		if (filter instanceof BufferedImageFilter2)
			filter = new BufferedImageFilter2AsImageFilter((BufferedImageFilter2) filter);
		
		if (filter instanceof ImageFilter) {
			return (ImageFilter) filter;
		} else {
			throw new IllegalArgumentException("unsupported filter type: "+filter.getClass().getName());
		}
	}
	
	// ### MODEL ###
	
	static class FilterComboBoxModel extends AbstractListModel implements ComboBoxModel {
		// ### ELEMENTS ###
		
		public Object getElementAt(int index) {
			return filterList.get(index);
		}
		
		public int getSize() {
			return filterList.size();
		}
		
		// ### COMBO BOX MODEL ###
		
		private Object selectedItem = null;
		
		public Object getSelectedItem() {
			return this.selectedItem;
		}
		
		public void setSelectedItem(Object selectedItem) {
			this.selectedItem = selectedItem;
		}
	}
}

class JDesktopMessagePane extends JLabel {
	public JDesktopMessagePane() {
		this.setHorizontalAlignment(SwingConstants.CENTER);
		this.setForeground(Color.WHITE);
		this.setBackground(Color.GRAY);
	}
	
	public JDesktopMessagePane(String text) {
		this();
		this.setText(text);
	}
	
	// ### IMPL ###
	
	public void paintComponent(Graphics g) {
		g.setColor(this.getBackground());
		g.fillRect(0, 0, this.getWidth(), this.getHeight());
		super.paintComponent(g);
	}
	
	// Ensure that the preferred size is always square
	public Dimension getPreferredSize() {
		Dimension prefSize = super.getPreferredSize();
		prefSize.width = prefSize.height = Math.max(prefSize.width, prefSize.height);
		return prefSize;
	}
	
	// ### ACCESSORS ###
	
	public void setText(String text) {
		super.setText(text);
	}
}