public class OtsuThresholdingAlgorithm {
	/*
	 * Calculates the optimal threshold level for a 2-level thresholding of the input image.
	 * The performance of this algorithm is estimated to be O(1).
	 * 
	 * This implementation could be generalized to calculate the (N-1) optimal
	 * thresholds for an N-level thresholding of the input image. It should be
	 * stressed, however, that a such implementation would probably be computationally
	 * intensive (I would estimate the performance to be O(256^N)). Even a
	 * clever implementation I believe would have a performance of around
	 * O(k*(256^(N-1) + ... + 256^1) + 256) for small k, which is
	 * asymptotically equivalent to O(256^(N-1)) which isn't much better
	 * than the O(256^N) performance of the naive algorithm.
	 */
	public static int calculateOptimalThreshold(int width, int height, int[] inGrayPixels) {
		int numPixels = inGrayPixels.length;
		
		int[] histogram;
		{
			histogram = new int[256];
			for (int curLevel=0; curLevel<256; curLevel++) {
				histogram[curLevel] = 0;
			}
			for (int i=0; i<numPixels; i++) {
				int curGrayPixel = inGrayPixels[i];
				int curLevel = curGrayPixel;
				histogram[curLevel]++;
			}
		}
		
		return calculateOptimalThresholdWithHistogram(histogram, numPixels);
	}
	
	public static int calculateOptimalThresholdWithHistogram(int[] histogram, int numPixels) {
		float[] probabilityHistogram;
		{
			probabilityHistogram = new float[256];
			
			float numPixelsAsFloat = (float) numPixels;
			for (int curLevel=0; curLevel<256; curLevel++) {
				probabilityHistogram[curLevel] = ((float) histogram[curLevel])/numPixelsAsFloat;
			}
		}
		
		/* Begin Otsu calculation */
		float C1_omega, C1_mu;
		float C2_omega, C2_mu;
		
		float fullMu = 				// mean intensity of the entire image
			calculateFirstMoment(
				probabilityHistogram,
				0, probabilityHistogram.length);
		
		/*
		 * Find the threshold value t such that the between-class variance
		 * (between class C1=[0,t] and C2=[t+1,<max>]) is maximized.
		 * 
		 * This is the optimal value t* for t.
		 */
		double maxSigmaSquared = 0;
		int thresholdWithMaxSigmaSquared = 0;
		for (int curThreshold=0; curThreshold<256; curThreshold++) {
			C1_omega =
				calculateMass(
					probabilityHistogram,
					0, curThreshold);
			C1_mu =
				calculateCenterOfMassWithMass(
					probabilityHistogram,
					0, curThreshold,
					C1_omega);
			
			C2_omega =
				calculateMass(
					probabilityHistogram,
					curThreshold+1, probabilityHistogram.length);
			C2_mu =
				calculateCenterOfMassWithMass(
					probabilityHistogram,
					curThreshold+1, probabilityHistogram.length,
					C2_omega);
			
			// between-class variance (between class C1 and C2)
			double curSigmaSquared = 
				C1_omega*square(C1_mu - fullMu) +
				C2_omega*square(C2_mu - fullMu);
			
			if (curSigmaSquared > maxSigmaSquared) {
				maxSigmaSquared = curSigmaSquared;
				thresholdWithMaxSigmaSquared = curThreshold;
			}
		}
		
		return thresholdWithMaxSigmaSquared;
	}
	
	// ### MASS, FIRST MOMENT, AND CENTER OF MASS CALCULATIONS ###
	
	// calculate mass AKA mean level
	private static float calculateMass(
		float[] probabilityHistogram, int startIndex, int limitIndex)
	{
		float mass = 0;
		for (int i=startIndex; i<limitIndex; i++) {
			mass += probabilityHistogram[i];
		}
		return mass;
	}
	
	// calculate first moment AKA mean intensity
	private static float calculateFirstMoment(
		float[] probabilityHistogram, int startIndex, int limitIndex)
	{
		float firstMoment = 0;
		for (int i=startIndex; i<limitIndex; i++) {
			firstMoment += i*probabilityHistogram[i];
		}
		return firstMoment;
	}
	
	private static float calculateCenterOfMassWithMass(
		float[] probabilityHistogram, int startIndex, int limitIndex, float mass)
	{
		return calculateCenterOfMassWithMassAndFirstMoment(
			mass,
			calculateFirstMoment(probabilityHistogram, startIndex, limitIndex));
	}
	
	private static float calculateCenterOfMassWithMassAndFirstMoment(
		float mass, float firstMoment)
	{
		return (mass != 0)
			? (firstMoment / mass)
			: 0;					// center of mass is undefined
	}
	
	// ### MISC ###
	
	private static float square(float n) {
		return n*n;
	}
}